class YamlFileNotFound(Exception):

    def __init__(self, yaml_path: str = ""):
        self.message = f"Path {yaml_path} not found."
        super().__init__(self.message)


class XPathValidationException(Exception):

    def __init__(self, field: str = ""):
        self.message = f"Invalid xpath expression for field: {field}"
        super().__init__(self.message)


class RegexValidationException(Exception):

    def __init__(self, field: str = ""):
        self.message = f"Invalid regex expression for field: {field}"
        super().__init__(self.message)



from typing import Dict

from pyparsy.enum_types import SelectorType, ReturnType


class Definition:

    __slots__ = (
        'field',
        'selector_type',
        'return_type',
        'multiple',
        'xpath',
        'regex',
        'children'

    )

    def __init__(self, field: str, definition: dict):
        self.field: str = field
        self.selector_type: SelectorType = SelectorType[definition.get("selector_type")]
        self.return_type: ReturnType = ReturnType[definition.get("return_type")]
        self.multiple: bool = definition.get("multiple", False)
        if self.selector_type.XPATH:
            self.xpath = definition.get('selector')
            self.regex = None
        else:
            self.xpath = None
            self.regex = definition.get("selector")
        self.children: Dict[str, Definition] = {
            _field: self.__class__(_field, _definitions) for _field, _definitions in definition.get("children", {}).items()
        }



import re
from decimal import Decimal

from typing import Optional


def __clean_string_number(given_string: str) -> str:
    """
    Format the given string to a numbered floating point string without using locales.

    :param given_string: string containing floating point number
    :return: str
    """
    number = re.sub(r"[^(\d,.)]", "", given_string)
    dot_position = number.find(".")
    comma_position = number.find(',')
    if comma_position > dot_position:
        number = number.replace('.', "")
        number = number.replace(",", ".")
    else:
        number = number.replace(",", "")
    return number


def extract_float(float_string: str) -> Optional[float]:
    """
    Extract floating point number from string

    :param float_string: string containing floating point number
    :return: float or None
    """
    if not float_string:
        return None
    return float(__clean_string_number(float_string))


def extract_decimal(decimal_string: str) -> Optional[Decimal]:
    """
    Extract Decimal from string

    :param decimal_string: string containing decimal
    :return: Decimal or None
    """
    if not decimal_string:
        return None
    return Decimal(__clean_string_number(decimal_string))


def extract_integer(integer_string: str) -> Optional[int]:
    """
    Extract single integer from string

    :param integer_string: string containing integer
    :return: int or None
    """
    if not integer_string:
        return None
    return int(re.sub(r"\D", "", integer_string))




import re
from typing import Dict, Union

import lxml.etree
from schema import Schema, And, Use, Optional, Or

from pyparsy.enum_types import SelectorType
from pyparsy.exceptions import XPathValidationException, RegexValidationException

DEFINITION_SCHEMA = Schema({
    And("selector"): Or(str, list[str]),
    And("selector_type"): And(str, Use(str.upper),
                              lambda s: s in ("XPATH", "REGEX")),
    Optional("multiple"): And(bool),
    And("return_type"): And(str, Use(str.upper),
                            lambda s: s in ("INTEGER", "STRING", "FLOAT", "DOUBLE", "MAP")),
    Optional("children"): Or(dict, list)
})


class Validator:

    def __init__(self, yaml_def: Dict):
        self.yaml_def = yaml_def
        self.__validate_all()
        self.is_valid = True

    def __validate_all(self):
        for field, definitions in self.yaml_def.items():
            self.validate_schema(definitions)
            if SelectorType[definitions.get("selector_type")] == SelectorType.XPATH:
                self.validate_xpath(definitions.get("selector"), field)
            if SelectorType[definitions.get("selector_type")] == SelectorType.REGEX:
                self.validate_regex(definitions.get("selector"), field)

    def validate_schema(self, definitions: Dict):
        try:
            DEFINITION_SCHEMA.validate(definitions)
        except Exception as e:
            raise e

    def validate_xpath(self, xpath: Union[str, list[str]], field: str):
        try:
            if isinstance(xpath, str):
                lxml.etree.XPath(xpath)
            else:
                for path in xpath:
                    lxml.etree.XPath(path)
        except lxml.etree.XPathSyntaxError:
            raise XPathValidationException(field)

    def validate_regex(self, regex: str, field: str):
        try:
            re.compile(regex)
        except Exception:
            raise RegexValidationException(field)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyparsy', 'pyparsy.exceptions', 'pyparsy.internal', 'pyparsy.validator']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.9.1,<5.0.0',
 'parsel>=1.7.0,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'schema>=0.7.5,<0.8.0']

setup_kwargs = {
    'name': 'pyparsy',
    'version': '0.1.1',
    'description': 'HTML parsing library using YAML definitions and XPath',
    'long_description': '![Logo](https://raw.githubusercontent.com/vkolev/parsy/master/images/parsy-logo.png)\n\n![CI](https://github.com/vkolev/parsy/actions/workflows/main.yml/badge.svg?branch=master) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyparsy) ![PyPI](https://img.shields.io/pypi/v/pyparsy)\n\n# PyParsy\n\nParsy is a HTML parsing library using YAML definition files. The idea is to use the YAML file as\nsort of intent - what you want to have as a result and let Parsy do the heavy lifting for you.\n\nThe YAML files contain:\n- The desired structure of the output\n- XPath variants of the parsed items\n\n## Features\n\n- YAML File definitions\n- Intent instead of coding\n- support for XPath and Regex\n- Different output formats e.g. JSON, YAML, XML\n\n## Installation\n\nUsing pip:\n```shell\npip install pyparsy\n```\n\n## Running Tests\n\nTo run tests, run the following command\n\n```bash\n  poetry run pytest\n```\n\n## Examples\n\nWe can consider as an example the amazon bestseller page. First we define the .yaml definition file:\n\n```yaml\ntitle:\n  selector: //div[contains(@class, "_card-title_")]/h1/text()\n  selector_type: XPATH\n  return_type: STRING\npage:\n  selector: //ul[contains(@class, "a-pagination")]/li[@class="a-selected"]/a/text()\n  selector_type: XPATH\n  return_type: INTEGER\nproducts:\n  selector: //div[@id="gridItemRoot"]\n  selector_type: XPATH\n  multiple: true\n  return_type: MAP\n  children:\n    image:\n      selector: //img[contains(@class, "a-dynamic-image")]/@src\n      selector_type: XPATH\n      return_type: STRING\n    title:\n      selector: //a[@class="a-link-normal"]/span/div/text()\n      selector_type: XPATH\n      return_type: STRING\n    price:\n      selector: //span[contains(@class, "a-color-price")]/span/text()\n      selector_type: XPATH\n      return_type: FLOAT\n    asin:\n      selector: //div[contains(@class, "sc-uncoverable-faceout")]/@id\n      selector_type: XPATH\n      return_type: STRING\n    reviews_count:\n      selector: //div[contains(@class, "sc-uncoverable-faceout")]/div/div/a/span/text()\n      selector_type: XPATH\n      return_type: INTEGER\n```\n\nFor the example sake let\'s store the file as `amazon_bestseller.yaml`.\n\nThen we can use the PyParsy library in out code:\n\n```python\nimport httpx\nfrom pyparsy import Parsy\n\ndef main():\n    html = httpx.get("https://www.amazon.com/gp/bestsellers/hi/?ie=UTF8&ref_=sv_hg_1")\n    parser = Parsy("amazon_bestseller.yaml")\n    result = parser.parse(html.text)\n    print(result)\n    \nif __name__ == "__main__":\n    main()\n```\n\nFor more examples please see the tests for the library.\n\n## Documentation\n\n[Documentation](https://parsy.readthedocs.com) (hopefuly some day)\n\n## Acknowledgements\n\n - [selectorlib](https://selectorlib.com/) - It is the main inspiration for this project\n - [Scrapy](https://scrapy.org/) - One of the best crawling libraries for Python\n - [Tiangolo](https://tiangolo.com/projects) - His projects are real inspiration to produce great software\n\n\n## Contributing\n\n',
    'author': 'Vladimir Kolev',
    'author_email': 'vladimir.r.kolev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/vkolev/parsy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from enum import Enum


class ReturnType(Enum):
    STRING = str
    INTEGER = int
    FLOAT = float
    BOOLEAN = bool
    MAP = dict


class SelectorType(Enum):
    XPATH = "xpath"
    REGEX = "regex"



import os
from collections import defaultdict
from typing import Any, Union

import yaml
from parsel import Selector, SelectorList
from yaml import SafeLoader

from pyparsy.exceptions import YamlFileNotFound
from pyparsy.enum_types import ReturnType, SelectorType
from pyparsy.internal import Definition
from pyparsy.utils import extract_float, extract_integer
from pyparsy.validator import Validator


class Parsy:

    def __init__(self, yaml_path: str = None, validate: bool = True):
        """
        Parsing class initializer

        :param yaml_path: path to YAML file containing definitions
        :param validate: bool - turn on/off yaml schema validation
        """
        self.yaml_path = yaml_path
        self._definitions = self.__load_definitions()
        if validate:
            self.__validate()
        self.field_selectors = self.__create_field_selectors()
        self.html_string = None

    def __load_definitions(self):
        if not os.path.exists(self.yaml_path):
            raise YamlFileNotFound(self.yaml_path)
        with open(self.yaml_path) as yaml_file:
            data = yaml.load(yaml_file, Loader=SafeLoader)
            return data

    def __create_field_selectors(self):
        if self._definitions:
            result = {}
            for field, definitions in self._definitions.items():
                result[field] = Definition(field, definitions)
            return result

    def __validate(self):
        try:
            return Validator(self._definitions)
        except Exception as e:
            raise e

    def parse(self, html_string: str):
        """
        Parse the whole html_string to a defaultdict

        :param html_string: HTML formated string
        :return: defaultdict()
        """
        result = defaultdict()
        html_data = Selector(html_string)
        for field, definition in self.field_selectors.items():
            if not definition.multiple:
                result[field] = self.parse_field(html_data, definition)
            else:
                result[field] = list(self.parse_filed_multiple(html_data, definition))
        return result

    def parse_field(self, html_data: Union[Selector, SelectorList], definition: Definition) -> Any:
        """
        Extract field from html_data with given definition
        :param html_data: parsel.Selector with HTML data
        :param definition: pyparsy.internal.Definition
        :return: Any
        """
        if definition.return_type != ReturnType.MAP:
            data = self._get_selector_data(html_data, definition)
            return self._convert_to_type(data, return_type=definition.return_type)
        result = defaultdict()
        data = self._get_selector_data(html_data, definition)
        for child, child_definition in definition.children.items():
            result[child] = self.parse_field(data, child_definition)
        return result

    def parse_filed_multiple(self, html_data: Union[Selector, SelectorList], definition) -> Any:
        items = self._get_selector_data(html_data, definition).getall()
        for item in items:
            html_data = Selector(item)
            result = defaultdict()
            for field, field_definition in definition.children.items():
                result[field] = self.parse_field(html_data, field_definition)
            yield result

    def _get_selector_data(self, html_data: Selector, definition: Definition):
        if definition.selector_type == SelectorType.XPATH:
            return html_data.xpath(query=definition.xpath)
        return html_data.re(definition.regex)

    def _convert_to_type(self, html_data: Union[Selector, SelectorList], return_type: ReturnType):
        if return_type == ReturnType.STRING:
            return html_data.get()
        if return_type == ReturnType.INTEGER:
            return extract_integer(html_data.get())
        if return_type == ReturnType.FLOAT:
            return extract_float(html_data.get())
        if return_type == ReturnType.BOOLEAN:
            return html_data.get() is not None
        return None



